#include <stdio.h>

#include "core/utils/utils.h"
#include "core/debug/evemon.h"
#include "core/storage/storage.h"
#include "core/storage/buffer.h"
#include "core/storage/fs/object.h"
#include "core/storage/file.h"
#include "core/storage/base.h"
#include "core/types/endian.h"
#include "core/types/type.h"

#include "sample.h"

// display help
void help(void)
{
	printf("Usage:\r\n");
	printf("\tspc dump \"filename\"\r\n");
	printf("\r\n");
}

// recursive print of storage objects
void rprint(const std::shared_ptr<const storage::fs::object_base> obj, bool recursive, int depth = 0)
{
	if (obj == nullptr)
		return;

	// print current
	for (int i = 0; i < depth; i++)
		printf("   |");

	printf("   + %s %s::%s\r\n", obj->get_type_name().c_str(), obj->get_owner_name().c_str(), obj->get_var_name().c_str());

	// skip rest if non-recursive
	if (!recursive)
		return;

	// print vars
	auto vars = obj->get_vars();

	for (auto& p : vars)
	{
		if (p == nullptr)
			continue;

		for (int i = 0; i < depth + 1; i++)
			printf("   |");

		printf("   + %s %s::%s\r\n", p->type.c_str(), p->owner.c_str(), p->name.c_str());
	}

	// print sub-objects
	auto children = obj->get_children();

	for (auto& p : children)
		rprint(p, recursive, depth + 1);
}

// dump containers of file
int dump(int argc, char** argv)
{
	if (argc < 1)
		return 1;

	const char* filename = argv[0];

	// load file
	printf("Reading file %s...\r\n", filename);

	storage::container container;

	auto fs = container.unpack_ex(storage::load_buffer_from_file(filename));
	auto list = fs->get_list();

	printf("\r\n");

	// get number of containers
	if (list.size() == 0)
		printf("file is empty\r\n");
	else
	{
		printf("file has %zu container(s):\r\n", list.size());

		for (auto& v : list)
		{
			// convert to object
			auto obj = fs->create_object();

			obj->unpack(v);

			// print
			rprint(obj, true);
		}
	}

	return 0;
}

// create sample file
int sample(int argc, char** argv)
{
	if (argc < 1)
		return 1;

	const char* filename = argv[0];

	// simple instancer
	auto _create = [](const std::string& type)
	{
		if (is_type_name<sample_derived>(type))
			return std::static_pointer_cast<storage::base>(std::make_shared<sample_derived>());

		throw_exception(unknown_typename_exception, type);
	};

	// part 1
	{
		// create sample
		int a = rand() % 10;
		int b = rand() % 10;
		int c = rand() % 10;

		printf("Writing values (%d,%d,%d) to file %s...\r\n", a, b, c, filename);

		auto obj = std::make_shared<sample_derived>(a, b, c);

		// write file to disk
		storage::container container;

		try
		{
			// try to load file if existing
			container.unpack(storage::load_buffer_from_file(filename), _create);
		}
		catch (...) {}

		container.push_back("", "obj", obj);
		storage::save_buffer_to_file(filename, container.pack('SPC0'));

		printf("done.\r\n");
	}

	// part 2
	{
		printf("Loading sample file from disk...\r\n");

		storage::container container;

		container.unpack(storage::load_buffer_from_file(filename), _create);

		auto list = container.get_by_type(get_name_from_type<sample_derived>());

		for (auto& p : list)
		{
			auto pc = std::dynamic_pointer_cast<sample_derived>(p);

			if (pc == nullptr)
				continue;

			printf("file contains sample class with values (%d,%d,%d)\r\n", pc->m_test, pc->sample_base::m_test, pc->m_obj.m_test);
		}

		printf("done.\r\n");
	}


	// return successfuly
	return 0;
}

// main program
int main(int argc, char** argv)
{
	try
	{
		srand((unsigned int)time(NULL));

		if (argc < 2)
		{
			help();

			return 1;
		}

		if (_strcmpi(argv[1], "dump") == 0)
			return dump(argc - 2, argv + 2);
		else if (_strcmpi(argv[1], "sample") == 0)
			return sample(argc - 2, argv + 2);

		printf("Unknown command \"%s\"\r\n", argv[1]);

		help();
	}
	catch (exception::base& e)
	{
		printf("error: %s\r\n", e.to_string().c_str());

		_critical("%s", e.to_string().c_str());
	}
	catch (std::exception& e)
	{
		printf("error: %s\r\n", e.what());

		_critical("%s", e.what());
	}
	catch (...)
	{
		printf("unknown error!\r\n");

		_critical("unknown error!");
	}

	// signal error
	return 1;
}